package store

import (
	"encoding/json"
	"fmt"
	"github.com/bwmarrin/snowflake"
	log "github.com/sirupsen/logrus"
	"github.com/syndtr/goleveldb/leveldb"
	"os"
	"runtime"
	"strconv"
	"sync"
)

const revision = 1

var (
	imageNode *snowflake.Node
	userNode  *snowflake.Node
)

// Info represents system information of a store.
type Info struct {
	Revision       int         `json:"revision"`
	Compat         bool        `json:"compat"`
	Register       bool        `json:"register"`
	InitialUser    string      `json:"initial_user"`
	PermissionDir  os.FileMode `json:"permission_dir"`
	PermissionFile os.FileMode `json:"permission_file"`
}

// Store represents a file store.
type Store struct {
	Path           string
	SingleUser     bool
	Revision       int
	Compat         bool
	Register       bool
	InitialUser    string
	PermissionDir  os.FileMode
	PermissionFile os.FileMode
	pageldb        map[string]*leveldb.DB
	mutex          map[string]*sync.RWMutex
	sync.RWMutex
}

func init() {
	// Set Epoch to beginning of time (01/01/1970)
	snowflake.Epoch = 0

	// Create a single node
	var err error
	imageNode, err = snowflake.NewNode(9)
	userNode, err = snowflake.NewNode(7)
	if err != nil {
		log.Fatalf("Error creating snowflake generation node, %s", err)
	}
}

// New initialises a new store instance.
func New(path string, single bool) *Store {
	var store *Store
	if stat, err := os.Stat(path); err != nil {
		log.Infof("Initializing new store at %s.", path)
		// Initialise empty store.
		store = &Store{
			Path:           path,
			SingleUser:     single,
			Revision:       revision,
			Compat:         runtime.GOOS == "windows",
			Register:       false,
			PermissionDir:  0700,
			PermissionFile: 0600,
			pageldb:        make(map[string]*leveldb.DB),
			mutex:          make(map[string]*sync.RWMutex),
		}
		if err = store.create(); err != nil {
			log.Fatalf("Error creating store, %s", err)
		}
	} else {
		// Exit if store is not a directory.
		if !stat.IsDir() {
			log.Fatal("Store is not a directory.")
		}

		// Load and parse store info.
		var info Info
		var payload []byte
		if payload, err = os.ReadFile(path + "/" + infoJson); err != nil {
			log.Fatalf("Error reading store information, %s", err)
		} else {
			if err = json.Unmarshal(payload, &info); err != nil {
				log.Fatalf("Error parsing store information, %s", err)
			}
		}
		if info.Revision != revision {
			log.Fatalf("Store format revision %v, expecting revision %v.", info.Revision, revision)
		}

		// Create store instance.
		store = &Store{
			Path:           path,
			SingleUser:     single,
			Revision:       info.Revision,
			Compat:         info.Compat,
			Register:       info.Register,
			InitialUser:    info.InitialUser,
			PermissionDir:  info.PermissionDir,
			PermissionFile: info.PermissionFile,
			pageldb:        make(map[string]*leveldb.DB),
			mutex:          make(map[string]*sync.RWMutex),
		}
	}

	// Handle store locking
	if store.file(store.LockPath()) {
		if pid, err := os.ReadFile(store.LockPath()); err != nil {
			log.Fatalf("Store locked, lock file unreadable, %s", err)
		} else {
			log.Fatalf("Store locked by process %s.", string(pid))
		}
	}
	if err := os.WriteFile(store.LockPath(), []byte(strconv.Itoa(os.Getpid())), store.PermissionFile); err != nil {
		log.Fatalf("Error locking store, %s", err)
	}

	return store
}

// Close closes the store.
func (s *Store) Close() {
	// Close all leveldb
	for variant, ldb := range s.pageldb {
		if err := ldb.Close(); err != nil {
			log.Errorf("Error closing leveldb variant %s, %s", variant, err)
		}
		log.Infof("Page variant %s closed.", variant)
	}

	// Unlock store
	if err := os.Remove(s.LockPath()); err != nil {
		log.Errorf("Error unlocking store, %s", err)
	}

	log.Info("Store closed.")
}

// fatalClose closes the store and as cleanly as possible and exits with a fatal message.
func (s *Store) fatalClose(message string) {
	s.Close()
	log.Fatal(message)
}

// create sets up the store directory if it does not exist.
func (s *Store) create() error {
	// Check if exists
	if _, err := os.Stat(s.Path); err == nil {
		return AlreadyExists
	}

	// Create directories
	if err := os.Mkdir(s.Path, s.PermissionDir); err != nil {
		return err
	}
	if err := os.Mkdir(s.TagsDir(), s.PermissionDir); err != nil {
		return err
	}
	if err := os.Mkdir(s.ImagesDir(), s.PermissionDir); err != nil {
		return err
	}
	if err := os.Mkdir(s.ImagesSnowflakeDir(), s.PermissionDir); err != nil {
		return err
	}
	if err := os.Mkdir(s.UsersDir(), s.PermissionDir); err != nil {
		return err
	}
	if err := os.Mkdir(s.SecretsDir(), s.PermissionDir); err != nil {
		return err
	}
	if err := os.Mkdir(s.UsernamesDir(), s.PermissionDir); err != nil {
		return err
	}
	if err := os.Mkdir(s.PageBaseDir(), s.PermissionDir); err != nil {
		return err
	}

	// Create initial user
	info := s.UserAdd("root", "initial", true)
	if info.Snowflake == "" {
		log.Fatal("Error creating initial user.")
	} else {
		log.Infof("Created initial user with username root and password initial.")
	}
	s.InitialUser = info.Snowflake

	// Create information file
	if payload, err := json.Marshal(Info{
		Revision:       s.Revision,
		Compat:         s.Compat,
		Register:       s.Register,
		InitialUser:    s.InitialUser,
		PermissionDir:  s.PermissionDir,
		PermissionFile: s.PermissionFile,
	}); err != nil {
		return err
	} else {
		if err = os.WriteFile(s.Path+"/"+infoJson, payload, s.PermissionFile); err != nil {
			return err
		}
	}

	return nil
}

// getLock returns a lock associated with a string.
func (s *Store) getLock(entry string) *sync.RWMutex {
	s.RLock()
	mutex, ok := s.mutex[entry]
	s.RUnlock()
	if !ok {
		s.Lock()
		mutex = &sync.RWMutex{}
		s.mutex[entry] = mutex
		s.Unlock()
	}
	return mutex
}

// file probes for the existence of specified entry on the filesystem.
func (s *Store) file(path string) bool {
	if _, err := os.Stat(path); err != nil {
		if os.IsNotExist(err) {
			return false
		} else {
			s.fatalClose(fmt.Sprintf("Error stat %s, %s", path, err))
		}
	}
	return true
}

// dir probes for the presence of specified directory on the filesystem.
func (s *Store) dir(path string) bool {
	if stat, err := os.Stat(path); err != nil {
		if os.IsNotExist(err) {
			return false
		} else {
			s.fatalClose(fmt.Sprintf("Error stat %s, %s", path, err))
		}
	} else {
		if !stat.IsDir() {
			s.fatalClose(fmt.Sprintf("Path %s is not a directory.", path))
		}
	}
	return true
}

// flake validates a snowflake string.
func (s *Store) flake(flake string) bool {
	if flake == "" {
		return false
	}
	_, err := strconv.Atoi(flake)
	return err == nil
}

// link provides symlink-like usage while considering compat mode.
func (s *Store) link(old, new string) {
	if !s.Compat {
		if err := os.Symlink(old, new); err != nil {
			s.fatalClose(fmt.Sprintf("Error linking %s to %s, %s", old, new, err))
		}
	} else {
		if err := os.WriteFile(new, []byte(old), s.PermissionFile); err != nil {
			s.fatalClose(fmt.Sprintf("Error writing associate file for %s to %s, %s", old, new, err))
		}
	}
}

// readlink provides readlink-like usage while considering compat mode.
func (s *Store) readlink(path string) string {
	if !s.Compat {
		//if final, err := os.Readlink(path); err != nil {
		//	if os.IsNotExist(err) {
		//		return ""
		//	} else {
		//		log.Fatalf("Error reading symlink %s, %s.", path, err)
		//	}
		//} else {
		//	return final
		//}
		return path
	} else {
		if final, err := os.ReadFile(path); err != nil {
			if os.IsNotExist(err) {
				return ""
			} else {
				s.fatalClose(fmt.Sprintf("Error reading association file %s, %s.", path, err))
			}
		} else {
			return string(final)
		}
	}
	return ""
}
